"""User-configurable routing."""

from __future__ import annotations

import re
import json
import uuid
import random
import enum
import time
from collections import OrderedDict

import flask

from mock import RouteConfigurationError
from mock.auth import Auth
from mock.input import IncommingRequest

from typing import Optional, Dict, Any, List, Iterable


class Delay:
    """Delay allows you to se arbitrary time to wait before response is returned."""

    def __init__(self, min_delay: float = 0.0, max_delay: float = 0.0):
        if min_delay > max_delay:
            raise RouteConfigurationError(
                f'Minimum delay cannot be longer than maximum delay: {min_delay} > {max_delay}'
            )
        self.min_delay = min_delay
        self.max_delay = max_delay

    def serialize(self) -> Optional[List[float]]:
        """Convert Delay to json."""
        if self.min_delay == self.max_delay == 0.0:
            return None
        return [self.min_delay, self.max_delay]

    def wait(self) -> None:
        """Put program to sleep for random amount of time withing the specified range."""
        wait_time = random.uniform(self.min_delay, self.max_delay)
        time.sleep(wait_time)

    @classmethod
    def deserialize(cls, data: List[float]) -> Delay:
        """Convert json to Delay."""
        if data:
            return cls(*data)
        return cls()


class ResponseSelectionStrategy(enum.Enum):
    """Strategy of how to select a Response from list of responses."""

    cycle = 'cycle'
    random = 'random'
    greedy = 'greedy'

    def select_response_cycle(self, responses: List[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses.

        Consumes responses in order of definition. Cycles through items one by one.
        """
        candidate = None

        for response in responses:
            if response.is_active and (candidate is None or response.used_count < candidate.used_count):
                candidate = response
        return candidate

    def select_response_random(self, responses: List[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses.

        Selects random response from all available.
        """
        population = [r for r in responses if r.is_active]
        weights = [r.weight for r in population]
        result = random.choices(population=population, weights=weights, k=1)
        return result[0] if result else None

    def select_response_greedy(self, responses: List[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses.

        Consumes responses in order of definition until the first one is exhausted,
        then starts consuming the next in the row.
        """
        for response in responses:
            if response.is_active:
                return response
        return None

    def select_response(self, responses: Iterable[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses."""
        method_name = f'select_response_{self.value}'
        method = getattr(self, method_name)
        return method(responses)

    def serialize(self) -> str:
        """Convert ResponseSelectionStrategy to json."""
        return self.value

    @classmethod
    def deserialize(cls, method: str) -> ResponseSelectionStrategy:
        """Convert json to ResponseSelectionStrategy."""
        return cls(method or 'greedy')


class Response:
    """Container for predefined response."""

    def __init__(
        self,
        id: str,
        delay: Delay,
        body: Any = "",
        headers: Optional[Dict[str, Any]] = None,
        status: int = 200,
        repeat: Optional[int] = None,
        weight: float = 0.5
    ) -> None:
        self.id = id
        self.body = body
        self.headers = headers or {}
        self.status = status
        self.repeat = repeat
        self.weight = weight
        self.delay = delay
        self.used_count = 0

    @property
    def serialized_body(self) -> str:
        """Convert specified response body to string."""
        if isinstance(self.body, str):
            return self.body
        else:
            return json.dumps(self.body)

    def as_flask_response(self) -> flask.Response:
        """Convert Request to flask.Response suitable to return from an endpoint."""
        return flask.Response(
            response=self.serialized_body,
            status=self.status,
            headers=self.headers
        )

    def serialize(self) -> Dict[str, Any]:
        """Convert Response to json."""
        return {
            'id': self.id,
            'is_active': self.is_active,
            'status': self.status,
            'used_count': self.used_count,
            'weight': self.weight,
            'headers': self.headers,
            'repeat': self.repeat,
            'delay': self.delay.serialize(),
            'body': self.body
        }

    def use(self) -> None:
        """Increases usage counter of Response."""
        self.used_count += 1

    @property
    def is_active(self) -> bool:
        """Return True if response has some uses left."""
        return self.repeat is None or self.repeat > self.used_count

    def wait(self) -> None:
        """Sleep for time specified in the response."""
        self.delay.wait()

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Response:
        """Convert json to Response."""
        delay = Delay.deserialize(data.pop('delay', None))
        return cls(delay=delay, **data)


class Route:
    """Route is a pair of request arguments and all possible reponses."""

    def __init__(
        self,
        id: str,
        responses: OrderedDict[str, Response],
        response_selection: ResponseSelectionStrategy,
        path: re.Pattern,
        auth: Auth,
        method: str = 'GET'
    ):
        self.id = id
        self.responses = responses
        self.response_selection = response_selection
        self.method = method
        self.path = path
        self.auth = auth
        self.used_count = 0

    def serialize(self) -> Dict[str, Any]:
        """Convert Route to JSON."""
        return {
            'id': self.id,
            'response_selection': self.response_selection.serialize(),
            'auth': self.auth.serialize(),
            'method': self.method,
            'path': self.path.pattern,
            'used_count': self.used_count,
            'responses': [response.serialize() for response in self.responses.values()]
        }

    def get_response(self, response_id: str) -> Optional[Response]:
        """Get a Response by its id."""
        for response in self.responses.values():
            if response.id == response_id:
                return response
        return None

    @classmethod
    def _create_responses(cls, responses: List[Dict[str, Any]]) -> OrderedDict[str, Response]:
        """Create Response objects from json."""
        result = OrderedDict()
        for response in responses:
            if 'id' not in response:
                response['id'] = str(uuid.uuid4())
            elif response['id'] in result:
                raise RouteConfigurationError(f'Duplicate response id {response["id"]}.')

            result[response['id']] = Response.deserialize(response)
        return result

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Route:
        """Convert json to Route."""
        id = data.pop('id')
        auth = Auth.deserialize(data.pop('auth', None))
        path = re.compile(data.pop('path', None))
        response_selection = ResponseSelectionStrategy.deserialize(data.pop('response_selection', None))
        responses = cls._create_responses(data.pop('responses'))

        return cls(
            id=id,
            responses=responses,
            response_selection=response_selection,
            path=path,
            auth=auth,
            **data
        )

    def use(self, response: Optional[Response]) -> None:
        """Increment use counter of this Route and given Response."""
        self.used_count += 1
        if response:
            response.use()

    def match(self, request: IncommingRequest) -> bool:
        """Return True, if this request specification matches given InputRequest."""
        return self._match_method(request.method) and self._match_path(request.path)

    def _match_method(self, method: Optional[str]) -> bool:
        """Return True, if this requests HTTP method matches given InputRequest."""
        return self.method in [None, method]

    def _match_path(self, path: str) -> bool:
        """Return True, if this requests path matches given InputRequest."""
        return bool(self.path.match(path))

    def select_response(self) -> Optional[Response]:
        """Select response from list of responses."""
        return self.response_selection.select_response(self.responses.values())

    def authenticate(self, request: IncommingRequest) -> None:
        """Check if Request if properly authenticated."""
        self.auth.authenticate(request)


class Router:
    """Custom request/response router."""

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Remove all custom routes."""
        self.routes: OrderedDict[str, Route] = OrderedDict()

    def _validate_route_ids(self, route_ids: List[Optional[str]]) -> None:
        """Check that there are no duplicate route id."""
        for route_id in route_ids:
            if route_id is None:
                continue
            if route_id in self.routes:
                raise RouteConfigurationError(f'Route id {route_id} already exists.')
            if route_ids.count(route_id):
                raise RouteConfigurationError(f'Duplicate oute id {route_id}.')

    def _generate_route_id(self) -> str:
        """Generate route id."""
        while True:
            route_id = str(uuid.uuid4())
            if route_id not in self.routes:
                return route_id

    def _add_validated_route(self, route: Dict[str, Any]) -> Route:
        """Add route assuming its id was already validated."""
        if 'id' not in route:
            route['id'] = self._generate_route_id()

        new_route = Route.deserialize(route)
        self.routes[new_route.id] = new_route
        return new_route

    def add_route(self, route: Dict[str, Any]) -> Route:
        """Add custom request and matching responses."""
        self._validate_route_ids([route.get('id')])
        return self._add_validated_route(route)

    def add_routes(self, routes: List[Dict[str, Any]]) -> List[Route]:
        """Add multiple custom requests and matching responses."""
        self._validate_route_ids([route.get('id') for route in routes])
        result = []
        for route in routes:
            result.append(self._add_validated_route(route))
        return result

    def get_route(self, route_id: str) -> Optional[Route]:
        """Get Route by its id."""
        return self.routes.get(route_id, None)

    def match(self, incomming_request: IncommingRequest) -> Optional[Route]:
        """Find matching request and return apropriet response or None if none found."""
        for route in self.routes.values():
            if route.match(incomming_request):
                return route
        return None

    def serialize(self) -> List[Dict[str, Any]]:
        """Convert Router to dictionary suitable for sending as api response."""
        return [route.serialize() for route in self.routes.values()]



"""Install lmc-radlib and dependencies."""

import sys

from setuptools import setup


def main():
    setup(
        name='double',
        version='1.0.0',
        python_requires='>=3.6',
        description='Mocking Service',
        long_description='',
        url='https://github.com/JakubTesarek/double',
        author='Jakub Tesárek',
        author_email='jakub@tesarek.me',
        license='unlicensed',
        include_package_data=True,
        packages=['double'],
        install_requires=[
            'python-rapidjson'
        ],
        extras_require={
            'dev': [
                'setuptools',
                'wheel',
                'pytest',
                'pytest-cov',
                'pytest-mock',
                'mypy',
                'flake8',
                'flake8-colors',
                'flake8-eradicate',
                'flake8-print',
                'flake8-todo',
                'confluent-kafka',
                'kafka-python',
                'twine'
            ],
            'kafka': [
                'confluent-kafka'
            ]
        }
    )


if __name__ == '__main__':
    main()


"""Mock is a service providing REST API configurable by REST API."""


class MockException(Exception):
    """Exception from Mock service."""


class RouteConfigurationError(MockException, ValueError):
    """Raised when route could not be configured because of value error."""


class AuthenticationError(MockException):
    """Exception raised when user could not be authenticated."""



"""Incomming requests."""

from __future__ import annotations

import abc

import flask
import urllib.parse

from typing import Dict, Any


HTTP_METHODS = [
    'GET',
    'HEAD',
    'POST',
    'PUT',
    'DELETE',
    'CONNECT',
    'OPTIONS',
    'TRACE',
    'PATCH'
]


class IncommingRequest(abc.ABC):
    """Incomming request that can be validated or matched against routes."""

    @abc.abstractproperty
    def method(self) -> str:
        """HTTP method."""

    @abc.abstractproperty
    def path(self) -> str:
        """Path of the request: `http://domain.com/<path>?query`."""

    @abc.abstractproperty
    def headers(self) -> Dict[str, Any]:
        """Dictionary containing header name and value."""

    @abc.abstractproperty
    def args(self) -> Dict[str, Any]:
        """Dictionary containing URL arguments."""

    @abc.abstractproperty
    def url(self) -> str:
        """Full url of the request."""

    @abc.abstractproperty
    def query_string(self) -> bytes:
        """Query string of the request including ?: `http://domain.com/path<?query>`."""

    @abc.abstractproperty
    def form(self) -> Dict[str, Any]:
        """Dictionary containing form data."""

    @abc.abstractproperty
    def cookies(self) -> Dict[str, Any]:
        """Dictionary containing cookies."""


class IncommingFlaskRequest(IncommingRequest):
    """Request made by calling and API endpoint."""

    def __init__(self, request: flask.Request):
        self.request = request

    @property
    def method(self) -> str:
        """HTTP method."""
        return self.request.method

    @property
    def path(self) -> str:
        """Path of the request: `http://domain.com/<path>?query`."""
        return self.request.path

    @property
    def headers(self) -> Dict[str, Any]:
        """Dictionary containing headers."""
        return {key: value for key, value in self.request.headers.items()}

    @property
    def args(self) -> Dict[str, Any]:
        """Dictionary containing URL arguments."""
        return self.request.args

    @property
    def url(self) -> str:
        """Full url of the request."""
        return self.request.url

    @property
    def query_string(self) -> bytes:
        """Query string of the request including ?: `http://domain.com/path<?query>`."""
        return self.request.query_string

    @property
    def form(self) -> Dict[str, Any]:
        """Dictionary containing form data."""
        return self.request.form

    @property
    def cookies(self) -> Dict[str, Any]:
        """Dictionary containing cookies."""
        return self.request.cookies


class IncommingTestRequest:
    """Model of a request used for testing route matching."""

    def __init__(self, base_url: str, full_path: str, method: str):
        self.url = urllib.parse.urljoin(base_url, full_path)
        self.parsed_url = urllib.parse.urlparse(self.url)
        self.method = method

    @property
    def path(self) -> str:
        """Path of the request: `http://domain.com/<path>?query`."""
        return self.parsed_url.path

    @property
    def args(self) -> Dict[str, Any]:
        """Dictionary containing URL arguments."""
        args = urllib.parse.parse_qs(self.parsed_url.query)
        parsed_args = {}
        for key, values in args.items():
            parsed_args[key] = values[0] if len(values) == 1 else values
        return args

    @property
    def query_string(self) -> bytes:
        """Query string of the request including ?: `http://domain.com/path<?query>`."""
        return self.parsed_url.query.encode('utf-8')

    @property
    def headers(self) -> Dict[str, Any]:
        """Dictionary containing headers. Always empty."""
        return {}

    @property
    def form(self) -> Dict[str, Any]:
        """Dictionary containing form data. Always empty."""
        return {}

    @property
    def cookies(self) -> Dict[str, Any]:
        """Dictionary containing cookies. Always empty."""
        return {}



"""Authentication module."""

from __future__ import annotations

import abc
import re
import datetime
import hashlib
import urllib.parse
import hmac

import basicauth

from mock import RouteConfigurationError, AuthenticationError
from mock.input import IncommingRequest

from typing import Optional, Type, Any, Dict, Tuple, List


class Auth(abc.ABC):
    """Authentication class."""

    method: Optional[str] = None

    def __init__(self, *args: List, **kwargs: Dict):
        pass

    @abc.abstractmethod
    def authenticate(self, request: IncommingRequest) -> None:
        """Check if IncommingRequest contains valid authentication, raise exception if not."""

    def serialize(self) -> Dict[str, Any]:
        """Convert Auth to json value."""
        return {
            'method': self.method
        }

    @classmethod
    def _find_implementation(cls, method: str) -> Type[Auth]:
        """Find a subclass implementing given auth method."""
        for subclass in cls.__subclasses__():
            if subclass.method == method:
                return subclass
        raise RouteConfigurationError(f'Implementation of "{method}"" authentication method not found.')

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Auth:
        """Convert json value to Auth."""
        if data is None:
            return NoAuth()

        if 'method' in data:
            method = data.pop('method')
            implementation = cls._find_implementation(method)
        else:
            implementation = cls
        return implementation(**data)


class NoAuth(abc.ABC):
    """Placeholder authentication method. Doesn't perform authentication."""

    method = None

    def authenticate(self, request: IncommingRequest) -> None:
        """Authenticate IncommingRequest. Always succeds."""
        return None

    def serialize(self) -> None:
        """Convert Auth to json value."""
        return None


class TokenAuth(Auth):
    """Authentication using http token in header."""

    method = 'token'

    def __init__(self, token: str):
        self.token = token

    def _get_header(self, request: IncommingRequest) -> str:
        """Get value of http header containing authentication token."""
        header = request.headers.get('Authorization')
        if not header:
            raise AuthenticationError('Missing authentication header "Authorization".')
        return header

    def _get_token(self, header: str) -> str:
        """Get authetication token from http header."""
        match = re.match(r'Bearer (?P<token>.*)', header)
        if not match:
            raise AuthenticationError(f'Invalid authentication header {header}.')

        return match['token']

    def authenticate(self, request: IncommingRequest) -> None:
        """Check if IncommingRequest contains valid token authentication, raise exception if not."""
        header = self._get_header(request)
        token = self._get_token(header)
        if token != self.token:
            raise AuthenticationError(f'Authentication token {token} doens\'t match {self.token}.')

    def serialize(self) -> Dict[str, str]:
        """Convert TokenAuth to json value."""
        data = super().serialize()
        data.update({
            'token': self.token
        })
        return data


class BasicAuth(Auth):
    """Authentication using username and password in http header."""

    method = 'basic'

    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

    def _get_header(self, request: IncommingRequest) -> str:
        """Get string containing base64 string containting username and password."""
        token = request.headers.get('Authorization')
        if not token:
            raise AuthenticationError('Missing authentication header "Authorization".')
        return token

    def _get_username_password(self, token: str) -> Tuple[str, str]:
        """Parse base64 string to username and password."""
        try:
            return basicauth.decode(token)
        except basicauth.DecodeError:
            raise AuthenticationError(f'Invalid authentication header {token}.')

    def authenticate(self, request: IncommingRequest) -> None:
        """Check if IncommingRequest contains valid username and password, raise exception if not."""
        token = self._get_header(request)
        username, password = self._get_username_password(token)
        if username != self.username or password != self.password:
            raise AuthenticationError(
                f'Authentication {username}:{password} doens\'t match {self.username}:{self.password}.'
            )

    def serialize(self) -> Dict[str, str]:
        """Convert BasicAuth to json value."""
        data = super().serialize()
        data.update({
            'username': self.username,
            'password': self.password,
        })
        return data


class HmacAuth(Auth):
    """Authentication using hmac signature in url."""

    method = 'hmac'

    def __init__(self, key: str):
        self.key = key
        self.past_tolerance = 3600
        self.future_tolerance = 5

    def _hash_string(self, url: str) -> str:
        """Hash given URL using HMAC with SHA1 digest."""
        hash_maker = hmac.new(self.key.encode('utf-8'), digestmod=hashlib.sha1)
        hash_maker.update(url.encode('utf-8'))
        return hash_maker.hexdigest()

    def _get_timestamp(self, args: Dict[str, str]) -> datetime.datetime:
        """Get timestamp from url."""
        if 'hmac_timestamp' not in args:
            raise AuthenticationError(
                'HMAC authentication failed, URL is missing required parameter: "hmac_timestamp".'
            )
        return datetime.datetime.fromtimestamp(float(args['hmac_timestamp']))

    def _get_signature(self, args: Dict[str, str]) -> str:
        """Get hmac signature from url."""
        if 'hmac_sign' not in args:
            raise AuthenticationError('HMAC authentication failed, URL is missing a required parameter: "hmac_sign".')
        return args['hmac_sign']

    def _check_time(self, timestamp: datetime.datetime) -> None:
        """Check if given timestamp is within allowed bound."""
        now = datetime.datetime.now()

        if timestamp > now + datetime.timedelta(seconds=self.future_tolerance):
            raise AuthenticationError(
                f'HMAC authentication failed, URL contains hmac_timestamp '
                f'more than {self.future_tolerance} seconds in the future: {timestamp}'
            )

        if timestamp < now - datetime.timedelta(seconds=self.past_tolerance):
            raise AuthenticationError(
                f'HMAC authentication failed, URL contains hmac_timestamp '
                f'more than {self.past_tolerance} seconds in the past: {timestamp}'
            )

    def _hash_url(self, url: str, query: bytes) -> str:
        """Calculate hash of used url using the hmac key."""
        parsed_url = urllib.parse.urlparse(url)
        hashable_url = parsed_url.path + '?' + re.sub(r'&hmac_sign=.*$', '', query.decode('utf-8'))
        return self._hash_string(hashable_url)

    def _check_signature(self, url_hash: str, signature: str) -> None:
        """Check if signature matches expected hash."""
        if not url_hash or url_hash != signature:
            raise AuthenticationError('HMAC authentication failed, hash in URL parameter "hmac_sign" is invalid.')

    def authenticate(self, request: IncommingRequest) -> None:
        """Check if IncommingRequest contains valid authentication, raise exception if not."""
        timestamp = self._get_timestamp(request.args)
        signature = self._get_signature(request.args)
        url_hash = self._hash_url(request.url, request.query_string)

        self._check_signature(url_hash, signature)
        self._check_time(timestamp)

    def serialize(self) -> Dict[str, str]:
        """Convert Auth to json value."""
        data = super().serialize()
        data.update({
            'key': self.key
        })
        return data


class FormAuth(Auth):
    """Authentication using form data."""

    method = 'form'

    def __init__(self, fields: Dict[str, str]):
        self.fields = fields

    def _get_field(self, form: Dict[str, str], field: str) -> str:
        """Get value of form field."""
        if field not in form:
            raise AuthenticationError(f'Missing authentication field "{field}".')
        return form[field]

    def authenticate(self, request: IncommingRequest) -> None:
        """Check if IncommingRequest contains valid authentication, raise exception if not."""
        for field, value in self.fields.items():
            sent_value = self._get_field(request.form, field)
            if value != sent_value:
                raise AuthenticationError(
                    f'Incorrect value "{sent_value}" in field "{field}", expected "{value}".'
                )

    def serialize(self) -> Dict[str, str]:
        """Convert Auth to json value."""
        data = super().serialize()
        data.update({
            'fields': self.fields
        })
        return data


class CookieAuth(Auth):
    """Authentication using http cookie."""

    method = 'cookie'

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value

    def _get_cookie(self, cookies: Dict[str, str]) -> str:
        """Get value of cookie."""
        if self.name not in cookies:
            raise AuthenticationError(f'Missing authentication cookie "{self.name}".')
        return cookies[self.name]

    def authenticate(self, request: IncommingRequest) -> None:
        """Check if IncommingRequest contains valid authentication, raise exception if not."""
        sent_value = self._get_cookie(request.cookies)
        if self.value != sent_value:
            raise AuthenticationError(
                f'Incorrect value "{sent_value}" of cookie "{self.name}", expected "{self.value}".'
            )

    def serialize(self) -> Dict[str, str]:
        """Convert Auth to json value."""
        data = super().serialize()
        data.update({
            'name': self.name,
            'value': self.value
        })
        return data



"""API endpoints."""

from __future__ import annotations

from flask import Blueprint, Response, request, current_app, jsonify, abort

from mock.router import RouteConfigurationError
from mock.input import IncommingTestRequest, IncommingFlaskRequest, HTTP_METHODS
from mock.validation import request_schema
from mock.auth import AuthenticationError


internal_api = Blueprint('internal_api', __name__)
external_api = Blueprint('external_api', __name__)


@internal_api.route('/routes', methods=['DELETE'])
def reset_router() -> Response:
    """Reset router configuration."""
    current_app.user_router.reset()
    return Response(response='', status=200)


@internal_api.route('/route', methods=['POST'])
@request_schema('route.schema.json')
def add_route() -> Response:
    """Create new route."""
    try:
        route = current_app.user_router.add_route(request.get_json())
        return jsonify(route.serialize())
    except RouteConfigurationError as error:
        abort(400, str(error))


@internal_api.route('/routes', methods=['POST'])
@request_schema('routes.schema.json')
def add_routes() -> Response:
    """Create multiple new routes."""
    try:
        routes = current_app.user_router.add_routes(request.get_json())
        return jsonify([route.serialize() for route in routes])
    except RouteConfigurationError as error:
        abort(400, str(error))


@internal_api.route('/routes', methods=['GET'])
def get_all_routes() -> Response:
    """Get list of configured routes and responses."""
    return jsonify(current_app.user_router.serialize())


@internal_api.route('/route/<string:route_id>', methods=['GET'])
def get_route(route_id: str) -> Response:
    """Get single route by id."""
    if route := current_app.user_router.get_route(route_id):
        return jsonify(route.serialize())
    abort(404, f'Route id "{route_id}" does not exist.')


@internal_api.route('/routes/match', methods=['POST'])
def match_route() -> Response:
    """Match configured routes against given request."""
    payload = request.get_json()
    incomming_request = IncommingTestRequest(
        base_url=request.host_url,
        full_path=payload['path'],
        method=payload['method']
    )

    if route := current_app.user_router.match(incomming_request):
        return jsonify(route.serialize())
    abort(404, 'No route was matched.')


@internal_api.route('/route/<string:route_id>/response/<string:response_id>', methods=['GET'])
def get_response(route_id: str, response_id: str) -> Response:
    """Reset router configuration."""
    if route := current_app.user_router.get_route(route_id):
        if response := route.get_response(response_id):
            return jsonify(response.serialize())
        abort(404, f'Response id "{response_id}" does not exist in request id "{route_id}".')
    abort(404, f'Route id "{route_id}" does not exist.')


@external_api.route('/<path:path>', methods=HTTP_METHODS)
def respond(path: str) -> Response:
    """Match request againts defined routes and return appropriet response."""
    try:
        if route := current_app.user_router.match(IncommingFlaskRequest(request)):
            if response := route.select_response():
                response.wait()
                route.use(response)
                return response.as_flask_response()
        abort(404)
    except AuthenticationError as error:
        abort(401, str(error))
